"""This package provides a widget to edit the configuration of a camera."""

from ._editor import CameraConfigurationEditor, RectangularROIEditor

__all__ = ["CameraConfigurationEditor", "RectangularROIEditor"]
