"""Contains components to edit a digital time lane."""

from ._delegate import DigitalTimeLaneDelegate
from ._model import DigitalTimeLaneModel

__all__ = ["DigitalTimeLaneModel", "DigitalTimeLaneDelegate"]
