# Form implementation generated from reading ui file '.\graphplot\visualizer_creators_selector.ui'
#
# Created by: PyQt6 UI code generator 6.6.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtWidgets


class Ui_VisualizerCreatorSelector(object):
    def setupUi(self, VisualizerCreatorSelector):
        VisualizerCreatorSelector.setObjectName("VisualizerCreatorSelector")
        VisualizerCreatorSelector.resize(400, 300)
        VisualizerCreatorSelector.setLayoutDirection(
            QtCore.Qt.LayoutDirection.LeftToRight
        )
        self.verticalLayout = QtWidgets.QVBoxLayout(VisualizerCreatorSelector)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.dataLoaderLabel = QtWidgets.QLabel(parent=VisualizerCreatorSelector)
        self.dataLoaderLabel.setObjectName("dataLoaderLabel")
        self.formLayout.setWidget(
            0, QtWidgets.QFormLayout.ItemRole.LabelRole, self.dataLoaderLabel
        )
        self._visualizer_combo_box = QtWidgets.QComboBox(
            parent=VisualizerCreatorSelector
        )
        self._visualizer_combo_box.setObjectName("_visualizer_combo_box")
        self.formLayout.setWidget(
            0, QtWidgets.QFormLayout.ItemRole.FieldRole, self._visualizer_combo_box
        )
        self.verticalLayout.addLayout(self.formLayout)
        self._settings_group = QtWidgets.QGroupBox(parent=VisualizerCreatorSelector)
        self._settings_group.setObjectName("_settings_group")
        self.verticalLayout.addWidget(self._settings_group)
        spacerItem = QtWidgets.QSpacerItem(
            20,
            40,
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        self.verticalLayout.addItem(spacerItem)
        self._apply_button = QtWidgets.QPushButton(parent=VisualizerCreatorSelector)
        self._apply_button.setObjectName("_apply_button")
        self.verticalLayout.addWidget(self._apply_button)

        self.retranslateUi(VisualizerCreatorSelector)
        QtCore.QMetaObject.connectSlotsByName(VisualizerCreatorSelector)

    def retranslateUi(self, VisualizerCreatorSelector):
        _translate = QtCore.QCoreApplication.translate
        VisualizerCreatorSelector.setWindowTitle(
            _translate("VisualizerCreatorSelector", "Form")
        )
        self.dataLoaderLabel.setText(
            _translate("VisualizerCreatorSelector", "DataView")
        )
        self._settings_group.setTitle(
            _translate("VisualizerCreatorSelector", "Settings")
        )
        self._apply_button.setText(_translate("VisualizerCreatorSelector", "Apply"))
